/*****************************************************************************
 * C/C++ Headers
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/*****************************************************************************
 * Qt Headers
 *****************************************************************************/

#include <qvalidator.h>

/*****************************************************************************
 * Local Headers
 *****************************************************************************/

#include "dimensionfield.h"

/*****************************************************************************
 * Local QValidator subclass: RealValidator
 *****************************************************************************/

class RealValidator : public QValidator
{
   public:
      RealValidator();
      State validate(QString &s, int &) const;
};

RealValidator::RealValidator() :
   QValidator(NULL, "RealValidator")
{
}

QValidator::State RealValidator::validate(QString &s, int &) const
{
   // Some strings are 'valid', that is - not clearly invalid
   // but not 'acceptable' (clearly valid), handle these cases first
   
   const char *value = s.latin1();
   int i             = 0;

   for(i = 0; value[i]; i ++)
      if(!isspace(value[i]))
         break;

   if(value[i] == '\0')
   {
      // The string is either empty, or composed entirely
      // of whitespace

      return(Valid);
   }

   if(value[i] == '.')
   {
      // If this is the end of the string, or of all subsequent characters
      // are zeros ...

      const char *ptr = value + i + 1;

      if(*ptr == '\0' || strspn(ptr, "0") == strlen(ptr))
         return(Valid);
   }

   if(strspn(value + i, "0") == strlen(value + i))
   {
      // Zero or more whitespace characters followed by one or more '0'
      // characters

      return(Valid);
   }

   // At this point, we're either going to return 'Acceptable' or 'Invalid'.
   // We'll make that decision based upon:
   //    (a) if strtod accepts the string and
   //    (b) the string contains none of: [eE+-]

   char *ptr = NULL;

   (void)strtod(value, &ptr);

   if(*ptr != '\0')
   {
      // strtod could not consume the entire string, it is therefore
      // invalid

      return(Invalid);
   }

   // Here we're specifically disallowing:
   //    (a) all negative values
   //    (b) all values beginning with an explicit '+'
   //    (c) all values written in exponential form
   //
   // The only one of these that is strictly invalid for us is (a)
   // since we require our values to be > 0, but we want to restrict
   // that set of acceptable inputs to just standard numbers, no
   // leading '+' or exponential notation allowed.
   //
   // The function 'strpbrk' returns a pointer to the first occurrence
   // in its first argument of any character in its second argument so
   // whether or not it returns NULL tells us if any of [eE+-]
   // were present

   return(strpbrk(value, "eE+-") == NULL ? Acceptable : Invalid);
}

/*****************************************************************************
 * Definition of Class: DimensionField
 *****************************************************************************/

// Note that we default 'mValue' to 1.0, we cannot use 0.0 as that value
// isn't valid for this class.  The default value of 1.0 will only persist
// if the passed-in 'value' isn't positive

DimensionField::DimensionField(
   QWidget *parent,
   double value,
   const char *name,
   void *clientData,
   FilterFunction in,
   FilterFunction out
)  :
   QLineEdit(parent, name),
   mValue(1.0),
   mValidator(new RealValidator()),
   mClientData(clientData),
   mIn(in),
   mOut(out)
{
   setValue(value);
   setValidator(mValidator);

   // Wire up QLineEdit's textChanged signal to our slotHandleTextChanged
   // slot - inside that slot is where we'll decide whether or not to
   // update our internal value and emit our valueChanged signal

   QObject::connect(
      this,
      SIGNAL(textChanged(const QString &)),
      this,
      SLOT(slotHandleTextChanged(const QString &))
   );
}

DimensionField::~DimensionField()
{
   delete mValidator;
}

void DimensionField::setFilters(FilterFunction in, FilterFunction out)
{
   // Ensure our value gets updated appropriately ...

   double current = getValue();
   mIn            = in;
   mOut           = out;
   setValue(current);
}

double DimensionField::getValue()
{
   return(mOut(mClientData, mValue));
}

int DimensionField::setValue(double theValue)
{
   if(theValue <= 0.0)
      return(FALSE);

   // Set our internal value from 'theValue' but let it be filtered
   // first ...

   mValue = mIn(mClientData, theValue);

   char temp[255] = { '\0' };

   sprintf(temp, "%.2f", mValue);

   // Does 'temp' represent an integer? If so, we'll display it
   // without the decimal places

   char *charPtr = strchr(temp, '.');

   if(charPtr != NULL && strcmp(charPtr, ".00") == 0)
      *charPtr = '\0';

   // Since we're setting the value programmatically, temporarily
   // disable (then reenable) signals so that we don't emit
   // 'textChanged' in response to this call

   blockSignals(TRUE);
   setText(temp);
   blockSignals(FALSE);

   return(TRUE);
}

int DimensionField::isAcceptable()
{
   return(isAcceptable(text()));
}

void DimensionField::slotHandleTextChanged(const QString &s)
{
   // We only emit our 'valueChanged' signal if/when our
   // current value is a valid number

   if(isAcceptable(s))
   {
      // Update our internal 'mValue' to reflect the text

      mValue = atof(text().latin1());

      // Let the outside world know as well ...

      emit valueChanged(getValue());
   }
}

int DimensionField::isAcceptable(const QString &s)
{
   // Our validator's validate method doesn't change the passed-in
   // QString, but the method signature for validate accepts a non-const
   // QString so we have to cast 's' before calling validate

   int position = 0;

   return(
      validator()->validate((QString &)s, position) == QValidator::Acceptable
   );
}

double DimensionField::nullFilter(void *, double value)
{
   return(value);
}
