// C/C++ Headers

#include <iostream>
#include <stdio.h>
#include <stdlib.h>

// Qt Headers

#include <qapplication.h>
#include <qpushbutton.h>
#include <qgroupbox.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qcombobox.h>
#include <qvalidator.h>
#include <qmessagebox.h>

// Local Headers

#include "example3.h"
#include "dimensionfield.h"
#include "QALE.h"

// Local #defines, just to avoid duplication since we use these
// in three places

#define PIXEL_WIDTH_LABEL      "pixelWidthLabel"
#define PIXEL_HEIGHT_LABEL     "pixelHeightLabel"
#define PRINT_WIDTH_LABEL      "printWidthLabel"
#define PRINT_HEIGHT_LABEL     "printHeightLabel"
#define PRINT_RESOLUTION_LABEL "printResolutionLabel"

/*****************************************************************************
 * Dialog Layout Specification
 *****************************************************************************/

static const char *dialogLayout =
"ScaleDialog(0, 0, 0, 0, LT)"
"{"
"   define {"
"      margin = 10"
"      gap    = 5"
""
"      labelRight = w * (70 / 290)"
"      fieldLeft  = w * (80 / 290)"
"      fieldWidth = w * (70 / 290)"
"      unitsLeft  = w * (155 / 290)"
"      unitsWidth = w * (105 / 290)"
"   }"
""
"   cancel(dw, dh,  w - $margin, $margin + $gap + dh, RT)"
"   ok(cancel.w, dh, cancel.r, cancel.t - $gap, RB)"
""
"   pixelGroup("
"      (w - $margin - cancel.w) - 3 * $margin,"
"      .3 * h,"
"      $margin,"
"      1.5 * $margin,"
"      LT"
"   )"
"   {"
"      define"
"      {"
"         margin     = 5"
"         y1         = $margin + h / 3"
"         y2         = $margin + 2 * h / 3"
"      }"
""
"      " PIXEL_WIDTH_LABEL "  (dw, dh, $labelRight, $y1, RC)"
"      " PIXEL_HEIGHT_LABEL " (dw, dh, $labelRight, $y2, RC)"
""
"      pixelWidth  ($fieldWidth, dh, $fieldLeft, $y1, LC)"
"      pixelHeight ($fieldWidth, dh, $fieldLeft, $y2, LC)"
""
"      pixelWidthUnits  ($unitsWidth, dh,  $unitsLeft, $y1, LC)"
"      pixelHeightUnits ($unitsWidth, dh,  $unitsLeft, $y2, LC)"
"   }"
""
"   printGroup("
"      pixelGroup.w, .4 * h, $margin, pixelGroup.b + 1.5 * $margin, LT"
"   )"
"   {"
"      define"
"      {"
"         margin        = 5"
"         y1            = $margin + h / 4"
"         y2            = $margin + h / 2"
"         y3            = $margin + 3 * h / 4"
"      }"
""
"      " PRINT_WIDTH_LABEL "      (dw, dh, $labelRight, $y1, RC)"
"      " PRINT_HEIGHT_LABEL "     (dw, dh, $labelRight, $y2, RC)"
"      " PRINT_RESOLUTION_LABEL " (dw, dh, $labelRight, $y3, RC)"
""
"      printWidth      ($fieldWidth, dh, $fieldLeft, $y1, LC)"
"      printHeight     ($fieldWidth, dh, $fieldLeft, $y2, LC)"
"      printResolution ($fieldWidth, dh, $fieldLeft, $y3, LC)"
""
"      printWidthUnits      ($unitsWidth, dh, $unitsLeft, $y1, LC)"
"      printHeightUnits     ($unitsWidth, dh, $unitsLeft, $y2, LC)"
"      printResolutionUnits ($unitsWidth, dh, $unitsLeft, $y3, LC)"
"   }"
""
"   constrainProportions("
"      dw, dh, 2 * $margin, printGroup.b + (h - printGroup.b) / 2, LC"
"   )"
"}";

/*****************************************************************************
 * Local struct: Entry
 *****************************************************************************/

struct Entry
{
   const char *theString;
   FilterFunction in;
   FilterFunction out;
};

/*****************************************************************************
 * Local Function Prototypes
 *****************************************************************************/

static Entry *localGetPixelWidthUnitsEntries();
static Entry *localGetPixelHeightUnitsEntries();
static Entry *localGetPrintWidthUnitsEntries();
static Entry *localGetPrintHeightUnitsEntries();
static Entry *localGetPrintResolutionUnitsEntries();
static void localLoadComboBox(QComboBox *box, Entry *entries);

/*****************************************************************************
 * Local struct: ScaleDialogPrivate
 *****************************************************************************/

struct ScaleDialogPrivate
{
   ScaleDialogPrivate(
      double widthPixels,
      double heightPixels,
      double dpi,
      ScaleDialog *scaleDialog
   );

   ~ScaleDialogPrivate();

   double mWidthPixels;
   double mHeightPixels;

   DimensionField *mPixelWidth;
   DimensionField *mPixelHeight;

   DimensionField *mPrintWidth;
   DimensionField *mPrintHeight;
   DimensionField *mPrintResolution;

   QCheckBox *mConstrain;

   QALE *mQALE;
};

/*****************************************************************************
 * definition of: struct ScaleDialogPrivate
 *****************************************************************************/

ScaleDialogPrivate::ScaleDialogPrivate(
   double widthPixels,
   double heightPixels,
   double dpi,
   ScaleDialog *scaleDialog
) :
   mWidthPixels(widthPixels),
   mHeightPixels(heightPixels)
{
   // Though QALE readily handles dialog resizing, here we just fix
   // the size as we achieve no benefit from allowing resizing

   scaleDialog->setFixedSize(400, 300);

   // Set up our ok and cancel buttons

   QPushButton *ok     = new QPushButton("OK",     scaleDialog, "ok");
   QPushButton *cancel = new QPushButton("Cancel", scaleDialog, "cancel");

   QObject::connect(ok,     SIGNAL(clicked()), scaleDialog, SLOT(accept()));
   QObject::connect(cancel, SIGNAL(clicked()), scaleDialog, SLOT(reject()));

   // Now our pixel dimensions group box and associated field labels

   QGroupBox *pixelGroup =
      new QGroupBox("Pixel Dimensions", scaleDialog, "pixelGroup");

   (void)new QLabel("Width",  pixelGroup, PIXEL_WIDTH_LABEL);
   (void)new QLabel("Height", pixelGroup, PIXEL_HEIGHT_LABEL);

   // Now set up our pixel height/width fields

   mPixelWidth  = new DimensionField(
      pixelGroup, mWidthPixels, "pixelWidth", (void *)this
   );

   mPixelHeight = new DimensionField(
      pixelGroup, mHeightPixels, "pixelHeight", (void *)this
   );

   // Wire up their associated signals/slots

   QObject::connect(
      mPixelWidth,
      SIGNAL(valueChanged(double)),
      scaleDialog,
      SLOT(slotPixelWidthChanged(double))
   );

   QObject::connect(
      mPixelHeight,
      SIGNAL(valueChanged(double)),
      scaleDialog,
      SLOT(slotPixelHeightChanged(double))
   );

   // Now the units fields

   QComboBox *pixelWidthUnits =
      new QComboBox(pixelGroup, "pixelWidthUnits");

   QComboBox *pixelHeightUnits =
      new QComboBox(pixelGroup, "pixelHeightUnits");

   // Load the combo boxes

   localLoadComboBox(pixelWidthUnits, localGetPixelWidthUnitsEntries());
   localLoadComboBox(pixelHeightUnits, localGetPixelHeightUnitsEntries());

   // Connect signals/slots

   QObject::connect(
      pixelWidthUnits,
      SIGNAL(activated(int)),
      scaleDialog,
      SLOT(slotPixelWidthUnitsChanged(int))
   );

   QObject::connect(
      pixelHeightUnits,
      SIGNAL(activated(int)),
      scaleDialog,
      SLOT(slotPixelHeightUnitsChanged(int))
   );

   // Now on to the print size group box

   QGroupBox *printGroup =
      new QGroupBox("Print Size", scaleDialog, "printGroup");

   (void)new QLabel("Width",      printGroup, PRINT_WIDTH_LABEL);
   (void)new QLabel("Height",     printGroup, PRINT_HEIGHT_LABEL);
   (void)new QLabel("Resolution", printGroup, PRINT_RESOLUTION_LABEL);

   double widthInches  = mWidthPixels / dpi;
   double heightInches = mHeightPixels / dpi;

   mPrintWidth =
      new DimensionField(printGroup, widthInches, "printWidth", (void *)this);

   mPrintHeight =
      new DimensionField(printGroup, heightInches, "printHeight", (void *)this);

   mPrintResolution =
      new DimensionField(printGroup, dpi, "printResolution", (void *)this);

   // Connect signals/slots

   QObject::connect(
      mPrintWidth,
      SIGNAL(valueChanged(double)),
      scaleDialog,
      SLOT(slotPrintWidthChanged(double))
   );

   QObject::connect(
      mPrintHeight,
      SIGNAL(valueChanged(double)),
      scaleDialog,
      SLOT(slotPrintHeightChanged(double))
   );

   QObject::connect(
      mPrintResolution,
      SIGNAL(valueChanged(double)),
      scaleDialog,
      SLOT(slotPrintResolutionChanged(double))
   );

   // Now create/load the associated combo boxes

   QComboBox *printWidthUnits =
      new QComboBox(printGroup, "printWidthUnits");

   QComboBox *printHeightUnits =
      new QComboBox(printGroup, "printHeightUnits");

   QComboBox *printResolutionUnits =
      new QComboBox(printGroup, "printResolutionUnits");
 
   localLoadComboBox(printWidthUnits, localGetPrintWidthUnitsEntries());
   localLoadComboBox(printHeightUnits, localGetPrintHeightUnitsEntries());

   localLoadComboBox(
      printResolutionUnits, localGetPrintResolutionUnitsEntries()
   );

   // Connect signals/slots

   QObject::connect(
      printWidthUnits,
      SIGNAL(activated(int)),
      scaleDialog,
      SLOT(slotPrintWidthUnitsChanged(int))
   );

   QObject::connect(
      printHeightUnits,
      SIGNAL(activated(int)),
      scaleDialog,
      SLOT(slotPrintHeightUnitsChanged(int))
   );

   QObject::connect(
      printResolutionUnits,
      SIGNAL(activated(int)),
      scaleDialog,
      SLOT(slotPrintResolutionUnitsChanged(int))
   );

   mConstrain = new QCheckBox(
      "Constrain Proportions", scaleDialog, "constrainProportions"
   );

   mConstrain->setChecked(TRUE);

   // Now we've created all our widgets, layout the dialog ...

   mQALE = QALE::constructFromText(scaleDialog, dialogLayout);
}

ScaleDialogPrivate::~ScaleDialogPrivate()
{
   delete mQALE;
}

/*****************************************************************************
 * definition of: class ScaleDialog
 *****************************************************************************/

ScaleDialog::ScaleDialog(
   int widthPixels,
   int heightPixels,
   double dpi,
   QWidget *parent
) :
   QDialog(parent, "ScaleDialog", TRUE),
   mPrivate(0)
{
   mPrivate = new ScaleDialogPrivate(widthPixels, heightPixels, dpi, this);
}

ScaleDialog::~ScaleDialog()
{
   delete mPrivate;
}

int ScaleDialog::getPixelWidth()
{
   return((int)(mPrivate->mPixelWidth->getValue() + 0.5));
}

int ScaleDialog::getPixelHeight()
{
   return((int)(mPrivate->mPixelHeight->getValue() + 0.5));
}

int ScaleDialog::main(int argc, char **argv)
{
   QApplication theApplication(argc, argv);

   ScaleDialog scaleDialog(640, 480);

   using std::cout;
   using std::endl;

   cout << "User pressed: "
        << (scaleDialog.exec() ? "Ok" : "Cancel")
        << endl;

   cout << "pixelWidth  : " << scaleDialog.getPixelWidth() << endl;
   cout << "pixelHeight : " << scaleDialog.getPixelHeight() << endl;

   return(0);
}

void ScaleDialog::done(int value)
{
   int reallyDone = TRUE;

   if(value == 1)
   {
      // User pressed 'ok', we need to verify that all fields contain
      // valid data before we allow the dialog to be dismissed.
      //
      // This local struct (Pair) associates a DimensionField to be
      // examined for validity with the name of the label field next to it.

      struct Pair
      {
         DimensionField *field;
         char *labelName;
      };

      Pair pairs[] =
      {
         { mPrivate->mPixelWidth,      PIXEL_WIDTH_LABEL      },
         { mPrivate->mPixelHeight,     PIXEL_HEIGHT_LABEL     },
         { mPrivate->mPrintWidth,      PRINT_WIDTH_LABEL      },
         { mPrivate->mPrintHeight,     PRINT_HEIGHT_LABEL     },
         { mPrivate->mPrintResolution, PRINT_RESOLUTION_LABEL },
      };

      int size  = sizeof(pairs) / sizeof(pairs[0]);
      int first = -1;
      QString badFields;

      for(int i = 0; i < size; i ++)
      {
         if(!pairs[i].field->isAcceptable())
         {
            // If this is the first bad field we save off the index
            // so we can give it the focus, also we prepend '\n'
            // if this isn't the first bad field

            if(first == -1)
               first = i;

            else
               badFields.append("\n");

            // Look up the text associate with the named QLabel, this
            // way we get internationalization for free

            QObject *target = child(pairs[i].labelName);

            // Ensure we get back what we expect ...

            if(target == NULL || !target->isA("QLabel"))
            {
               // Programming error if we end up here

               QString internalError;

               internalError.sprintf(
                  "Internal Error [1], (file: %s, line: %d)",
                  __FILE__, __LINE__
               );

               badFields.append(internalError);
            }

            else
            {
               QLabel *label = (QLabel *)target;
               target        = label->parent();

               // Ensure we get back what we expect ...

               if(target == NULL || !target->isA("QGroupBox"))
               {
                  // Programming error if we end up here

                  QString internalError;

                  internalError.sprintf(
                     "Internal Error [2], (file: %s, line: %d)",
                     __FILE__, __LINE__
                  );

                  badFields.append(internalError);
               }

               else
               {
                  QGroupBox *box = (QGroupBox *)target;

                  badFields.append(box->title());
                  badFields.append(" - ");
                  badFields.append(label->text());
               }
            }
         }
      }

      if(first != -1)
      {
         QString message(
            "All fields must contain valid values > 0.\n"
            "Please correct the following field(s)\n\n"
         );

         message.append(badFields);

         QMessageBox::warning(
            this, "Invalid Field(s)", message, QMessageBox::Ok, 0
         );

         // Now put focus back on the first offender ...

         pairs[first].field->setFocus();

         reallyDone = FALSE;
      }
   }
   
   if(reallyDone)
      QDialog::done(value);
}

/*****************************************************************************
 * slots
 *****************************************************************************/

void ScaleDialog::slotPixelWidthChanged(double value)
{
   double ratio       = value / mPrivate->mWidthPixels;
   double heightValue = ratio * mPrivate->mHeightPixels;
   double dpi         = mPrivate->mPrintResolution->getValue();

   mPrivate->mPrintWidth->setValue(value / dpi);

   if(mPrivate->mConstrain->isChecked())
   {
      mPrivate->mPixelHeight->setValue(heightValue);
      mPrivate->mPrintHeight->setValue(heightValue / dpi);
   }
}

void ScaleDialog::slotPixelHeightChanged(double value)
{
   double ratio      = value / mPrivate->mHeightPixels;
   double widthValue = ratio * mPrivate->mWidthPixels;
   double dpi        = mPrivate->mPrintResolution->getValue();

   mPrivate->mPrintHeight->setValue(value / dpi);

   if(mPrivate->mConstrain->isChecked())
   {
      mPrivate->mPixelWidth->setValue(widthValue);
      mPrivate->mPrintWidth->setValue(widthValue / dpi);
   }
}

void ScaleDialog::slotPixelWidthUnitsChanged(int index)
{
   Entry *theEntry = localGetPixelWidthUnitsEntries() + index;

   // Update the filters appropriately

   mPrivate->mPixelWidth->setFilters(theEntry->in, theEntry->out);
}

void ScaleDialog::slotPixelHeightUnitsChanged(int index)
{
   Entry *theEntry = localGetPixelHeightUnitsEntries() + index;

   // Update the filters appropriately

   mPrivate->mPixelHeight->setFilters(theEntry->in, theEntry->out);
}

void ScaleDialog::slotPrintWidthChanged(double value)
{
   double dpi         = mPrivate->mPrintResolution->getValue();
   double pixelWidth  = value * dpi;
   double ratio       = pixelWidth / mPrivate->mWidthPixels;
   double pixelHeight = ratio * mPrivate->mHeightPixels;

   mPrivate->mPixelWidth->setValue(pixelWidth);

   if(mPrivate->mConstrain->isChecked())
   {
      mPrivate->mPixelHeight->setValue(pixelHeight);
      mPrivate->mPrintHeight->setValue(pixelHeight / dpi);
   }
}

void ScaleDialog::slotPrintHeightChanged(double value)
{
   double dpi         = mPrivate->mPrintResolution->getValue();
   double pixelHeight = value * dpi;
   double ratio       = pixelHeight / mPrivate->mHeightPixels;
   double pixelWidth  = ratio * mPrivate->mWidthPixels;

   mPrivate->mPixelHeight->setValue(pixelHeight);

   if(mPrivate->mConstrain->isChecked())
   {
      mPrivate->mPixelWidth->setValue(pixelWidth);
      mPrivate->mPrintWidth->setValue(pixelWidth / dpi);
   }
}

void ScaleDialog::slotPrintResolutionChanged(double value)
{
   double pixelWidth   = value * mPrivate->mPrintWidth->getValue();
   double pixelHeight  = value * mPrivate->mPrintHeight->getValue();

   mPrivate->mPixelWidth->setValue(pixelWidth);
   mPrivate->mPixelHeight->setValue(pixelHeight);
}

void ScaleDialog::slotPrintWidthUnitsChanged(int index)
{
   Entry *theEntry = localGetPrintWidthUnitsEntries() + index;

   // Update the filters appropriately

   mPrivate->mPrintWidth->setFilters(theEntry->in, theEntry->out);
}

void ScaleDialog::slotPrintHeightUnitsChanged(int index)
{
   Entry *theEntry = localGetPrintHeightUnitsEntries() + index;

   // Update the filters appropriately

   mPrivate->mPrintHeight->setFilters(theEntry->in, theEntry->out);
}

void ScaleDialog::slotPrintResolutionUnitsChanged(int index)
{
   Entry *theEntry = localGetPrintResolutionUnitsEntries() + index;

   // Update the filters appropriately

   mPrivate->mPrintResolution->setFilters(theEntry->in, theEntry->out);
}

/*****************************************************************************
 * Local Functions
 *****************************************************************************/

// Conversion routines

static double localPixelWidthToPercent(void *clientData, double value)
{
   ScaleDialogPrivate *p = (ScaleDialogPrivate *)clientData;

   return((value / p->mWidthPixels) * 100.0);
}

static double localPercentWidthToPixels(void *clientData, double value)
{
   ScaleDialogPrivate *p = (ScaleDialogPrivate *)clientData;

   return((value / 100.0) * p->mWidthPixels);
}

static double localPixelHeightToPercent(void *clientData, double value)
{
   ScaleDialogPrivate *p = (ScaleDialogPrivate *)clientData;

   return((value / p->mHeightPixels) * 100.0);
}

static double localPercentHeightToPixels(void *clientData, double value)
{
   ScaleDialogPrivate *p = (ScaleDialogPrivate *)clientData;

   return((value / 100.0) * p->mHeightPixels);
}

static double localInchesToCM(void *, double value)
{
   return(value * 25.4);
}

static double localCMToInches(void *, double value)
{
   return(value / 25.4);
}

// JFD: These functions' bodies are very odd, but this is what is necessary
//      to mimic the photoshop gui

static double localPrintWidthToPercent(void *clientData, double)
{
   ScaleDialogPrivate *p = (ScaleDialogPrivate *)clientData;

   return(
      localPixelWidthToPercent(clientData, p->mPixelWidth->getValue())
   );
}

static double localPercentToPrintWidth(void *clientData, double)
{
   ScaleDialogPrivate *p = (ScaleDialogPrivate *)clientData;

   return(
      p->mPixelWidth->getValue() / p->mPrintResolution->getValue()
   );
}

static double localPrintHeightToPercent(void *clientData, double)
{
   ScaleDialogPrivate *p = (ScaleDialogPrivate *)clientData;

   return(
      localPixelHeightToPercent(clientData, p->mPixelHeight->getValue())
   );
}

static double localPercentToPrintHeight(void *clientData, double)
{
   ScaleDialogPrivate *p = (ScaleDialogPrivate *)clientData;

   return(
      p->mPixelHeight->getValue() / p->mPrintResolution->getValue()
   );
}

// Routines to supply pulldown units contents and associated filter functions

static Entry *localGetPixelWidthUnitsEntries()
{
   static Entry entries[] =
   {
      { "pixels",  DimensionField::nullFilter, DimensionField::nullFilter },
      { "percent", localPixelWidthToPercent,   localPercentWidthToPixels  },
      { NULL,      NULL,                       NULL                       },
   };


   return(entries);
}

static Entry *localGetPixelHeightUnitsEntries()
{
   static Entry entries[] =
   {
      { "pixels",  DimensionField::nullFilter,  DimensionField::nullFilter  },
      { "percent", localPixelHeightToPercent,   localPercentHeightToPixels  },
      { NULL,      NULL,                        NULL                        },
   };


   return(entries);
}

static Entry *localGetPrintWidthUnitsEntries()
{
   static Entry entries[] =
   {
      { "inches",  DimensionField::nullFilter,  DimensionField::nullFilter  },
      { "cm",      localInchesToCM,             localCMToInches             },
      { "percent", localPrintWidthToPercent,    localPercentToPrintWidth    },
      { NULL,      NULL,                        NULL                        },
   };

   return(entries);
}

static Entry *localGetPrintHeightUnitsEntries()
{
   static Entry entries[] =
   {
      { "inches",  DimensionField::nullFilter, DimensionField::nullFilter   },
      { "cm",      localInchesToCM,            localCMToInches              },
      { "percent", localPrintHeightToPercent,  localPercentToPrintHeight    },
      { NULL,      NULL,                       NULL                         },
   };

   return(entries);
}

static Entry *localGetPrintResolutionUnitsEntries()
{
   static Entry entries[] =
   {
      { "pixels/in", DimensionField::nullFilter, DimensionField::nullFilter  },
      { "pixels/cm", localCMToInches,            localInchesToCM             },
      { NULL,        NULL,                       NULL                        },
   };

   return(entries);
}

static void localLoadComboBox(QComboBox *box, Entry *entries)
{
   for(int i = 0; entries[i].theString; i ++)
      box->insertItem(entries[i].theString);
}

// For unit testing

//#ifdef __MAIN__

int main(int argc, char **argv)
{
   return(ScaleDialog::main(argc, argv));

}  /* main */

//#endif
