// C/C++ headers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <string>
#include <sstream>

using std::ostringstream;
using std::string;

// QALE headers

#include "AttributeNode.h"
#include "cpp_assert.h"
#include "Visitor.h"
#include "Component.h"

/*****************************************************************************
 * Static read-only struct mapping between attributes and their associated
 * string representations
 *****************************************************************************/

typedef struct
{
   char *asString;
   AttributeNode::Attribute theAttribute;

} Mapping;

static Mapping map[] =
{
   { "l",  AttributeNode::_LEFT           },
   { "r",  AttributeNode::_RIGHT          },
   { "t",  AttributeNode::_TOP            },
   { "b",  AttributeNode::_BOTTOM         },
   { "x",  AttributeNode::_XCENTER        },
   { "y",  AttributeNode::_YCENTER        },
   { "w",  AttributeNode::_WIDTH          },
   { "h",  AttributeNode::_HEIGHT         },
   { "dw", AttributeNode::_DEFAULT_WIDTH  },
   { "dh", AttributeNode::_DEFAULT_HEIGHT },
};

/*****************************************************************************
 * Definition of class: AttributeNode
 *****************************************************************************/

AttributeNode::AttributeNode(
   const char *name,
   Attribute theAttribute
)
   : mQualifiedName(NULL),
     mAttribute(theAttribute),
     mComponent(NULL)
{
   strcpy(mQualifiedName = new char[strlen(name) + 1], name);
}

AttributeNode::~AttributeNode()
{
   delete [] mQualifiedName;
}

AttributeNode::Attribute AttributeNode::get() const
{
   return(mAttribute);
}

double AttributeNode::evaluate() const
{
   double result = 0.0;

   switch(mAttribute)
   {
      case _LEFT          : result = mComponent->getLeft();          break;
      case _RIGHT         : result = mComponent->getRight();         break;
      case _TOP           : result = mComponent->getTop();           break;
      case _BOTTOM        : result = mComponent->getBottom();        break;
      case _XCENTER       : result = mComponent->getX();             break;
      case _YCENTER       : result = mComponent->getY();             break;
      case _WIDTH         : result = mComponent->getWidth();         break;
      case _HEIGHT        : result = mComponent->getHeight();        break;
      case _DEFAULT_WIDTH : result = mComponent->getDefaultWidth();  break;
      case _DEFAULT_HEIGHT: result = mComponent->getDefaultHeight(); break;

      default:
      {
         // We should never be here

         CPP_ASSERT(0);
      }
      break;
   }

   return(result);
}

ostream &AttributeNode::print(ostream &out) const
{
   if(*mQualifiedName)
      out << mQualifiedName << ".";

   return(out << attributeAsString());
}

Node *AttributeNode::copy() const
{
   return(
      new AttributeNode(
         mQualifiedName,
         mAttribute
      )
   );
}

void AttributeNode::accept(Visitor &v) const
{
   v.visited((AttributeNode *)this);
}

char *AttributeNode::getQualifiedName() const
{
   return(mQualifiedName);
}

void AttributeNode::setComponent(Component *c)
{
   mComponent = c;
}

AttributeNode *AttributeNode::createFromQualifiedIdentifier(
   char *identifier, string &errMsg
)
{
   char *dotPtr = strrchr(identifier, '.');

   CPP_ASSERT(dotPtr != NULL);

   ++dotPtr;

   char *buffer = new char[strlen(dotPtr) + 1];
   strcpy(buffer, dotPtr);

   int i = 0;

   for(i = 0; buffer[i]; i ++)
      if(buffer[i] >= 'A' && buffer[i] <= 'Z')
         buffer[i] = ('a' - 'A') + buffer[i];

   Attribute theAttribute = _UNKNOWN;

   int size = sizeof(map) / sizeof(map[0]);

   for(i = 0; i < size; i ++)
   {
      if(strcmp(buffer, map[i].asString) == 0)
      {
         theAttribute = map[i].theAttribute;
         break;
      }
   }

   delete [] buffer;

   if(theAttribute == _UNKNOWN)
   {
      ostringstream o;

      o << "Uknown attribute: [" << dotPtr << "]";

      errMsg = o.str();
      return(NULL);
   }

   dotPtr[-1] = '\0';

   return(
      new AttributeNode(identifier, theAttribute)
   );
}

const char *AttributeNode::attributeAsString() const
{
   int size = sizeof(map) / sizeof(map[0]);

   for(int i = 0; i < size; i ++)
      if(map[i].theAttribute == mAttribute)
         return(map[i].asString);

   // Shouldn't ever be here

   CPP_ASSERT(0);

   return(NULL);
}
