// C/C++ headers

#include <string.h>
#include <stdio.h>

#include <string>
using std::string;

// Our headers

#include "QALE.h"
#include "CountingMessageReceiver.h"
#include "PrintingMessageReceiver.h"
#include "Widget.h"
#include "StringMap.h"
#include "ComponentSupplier.h"
#include "DefineVisitor.h"
#include "PropertyVisitor.h"
#include "AttributeVisitor.h"
#include "qale_parse.h"

// Qt headers

#include <qwidget.h>
#include <qobject.h>
#include <qobjectlist.h>

/*****************************************************************************
 * Prototypes for functions local to this module
 *****************************************************************************/

static StringMap *createMap(MessageReceiver &receiver, QWidget *root);

static void addToMap(StringMap *map, const char *lineage, QObjectList *list);

/*****************************************************************************
 * Definition of class: QALE
 *****************************************************************************/

QALE::QALE(QWidget *root, Widget *mainWidget, MessageReceiver *receiver)
   : mWidget(mainWidget), mStillborn(0)
{
   // Wrapping "receiver" inside a CountingMessageReceiver allows us
   // to know how many errors have been encountered as we go through
   // the various processing stages

   CountingMessageReceiver counter(*receiver);

   // First we need to create a StringMap that maps the fully qualified
   // name of each widget to its associated QWidget object

   StringMap *map = createMap(counter, root);

   if(counter.getCount() > 0)
   {
      // No point in proceeding

      mStillborn = 1;
   }

   else
   {
      ComponentSupplier supplier(counter, mainWidget, map);

      if(counter.getCount() > 0)
      {
         mStillborn = 1;
      }

      else
      {
         // Now we need to fixup our defines ...

         DefineVisitor defineVisitor(counter);
         mWidget->resolveDefines(defineVisitor);

         if(counter.getCount() > 0)
         {
            mStillborn = 1;
         }
         else
         {
            // Now we fixup our property references ...

            PropertyVisitor propertyVisitor(counter);
            mWidget->resolveProperties(propertyVisitor);

            if(counter.getCount() > 0)
            {
               mStillborn = 1;
            }
            else
            {
               // Lastly we fixup our attribute references ...

               AttributeVisitor attributeVisitor(counter);
               mWidget->resolveAttributes(attributeVisitor);

               if(counter.getCount() > 0)
               {
                  mStillborn = 1;
               }
            }
         }
      }
   }

   delete map;

   process();
}

QALE::QALE()
   : mWidget(NULL), mStillborn(1)
{
   // Zombie constructor, to be called only when there was a parse
   // error
}


QALE::~QALE()
{
   delete mWidget;
}

void QALE::process() const
{
   // Do nothing unless we were properly constructed

   if(!mStillborn)
   {
      mWidget->setSize();
      mWidget->setLocation();
   }
}

QALE *QALE::constructFromFile(
   QWidget *root,
   const char *fileName,
   MessageReceiver *receiver
)
{
   int deleteReceiver = 0;

   if(receiver == NULL)
   {
      receiver       = new PrintingMessageReceiver();
      deleteReceiver = 1;
   }

   QALE *ptr     = NULL;
   string errMsg = "";
   char *text    = qale_slurpFile(fileName, errMsg);

   if(!text)
   {
      // Create a stillborn QALE object using the "zombie constructor"
      // so the caller doesn't have to check for NULL

      ptr = new QALE();
      receiver->accept(errMsg.c_str());
   }
   else
   {
      ptr = QALE::constructFromText(root, text, receiver);
      delete [] text;
   }

   if(deleteReceiver)
      delete receiver;

   return(ptr);
}

QALE *QALE::constructFromText(
   QWidget *root,
   const char *text,
   MessageReceiver *receiver
)
{
   int deleteReceiver = 0;

   if(receiver == NULL)
   {
      receiver       = new PrintingMessageReceiver();
      deleteReceiver = 1;
   }

   QALE *ptr         = NULL;
   string errMsg     = "";
   int errLineNumber = 0;

   Widget *w = qale_createFromText(text, errMsg, &errLineNumber);

   if(!w)
   {
      // For the moment we don't do anything with errLineNumber

      receiver->accept(errMsg.c_str());

      // Create a stillborn QALE object using the "zombie constructor"
      // so the caller doesn't have to check for NULL

      ptr = new QALE();
   }
   else
   {
      ptr = new QALE(root, w, receiver);
   }

   if(deleteReceiver)
      delete receiver;

   return(ptr);
}

static StringMap *createMap(
   MessageReceiver &receiver,
   QWidget *root
)
{
   StringMap *map   = new StringMap();
   const char *name = root->name();

   // It is a requirement that the top-level widget have a name.
   // We return an empty list if this is not the case

   if(name && strcmp(name, "") != 0)
   {
      map->put(name, root);

      // Grab all children that are themselves QWidgets

      addToMap(map, name, root->queryList("QWidget"));
   }
   else
   {
      // Warn the caller, this is the sort of thing he/she would likely
      // be keen to know about

      receiver.accept("Root qwidget has no name");
   }

   return(map);
}

static void addToMap(StringMap *map, const char *lineage, QObjectList *list)
{
   // For each object in "list" that has a name, create an entry in
   // our map

   QObjectListIt iterator(*list);
   QObject *object = NULL;

   while((object = iterator.current()) != NULL)
   {
      QWidget *q       = (QWidget *)object;
      const char *name = q->name();

      if(name && strcmp(name, "") != 0)
      {
         // It has a name, add it to our list after constructing its
         // lineage appropriately

         char *buffer =
            new char[strlen(lineage) + strlen(name) + strlen(".") + 1];

         sprintf(buffer, "%s.%s", lineage, name);

         map->put(buffer, q);
         addToMap(map, buffer, q->queryList("QWidget"));

         delete [] buffer;
      }

      ++iterator;
   }

   delete list;
}
